"""
The items module.

A part is an element that has been placed
in an specific track. We also consider a
part some imaginary things that are 
fulfilling the track regions in which the
user didn't put any media, so we can 
interact with those 'empty' items to 
obtain 'empty' audio and video frames.
"""